@@ -4,9 +4,7 @@
 # This module is released under the MIT License:
 # http://www.opensource.org/licenses/mit-license.php
 import logging
-import Queue
 import socket
-import types
 import ssl
 
 from google.protobuf.text_format import MessageToString
@@ -16,45 +14,28 @@ from smarkets.errors import reraise
 from smarkets.lazy import LazyCall
 from smarkets.streaming_api import eto, seto
 from smarkets.streaming_api.exceptions import ConnectionError, SocketDisconnected
+from smarkets.streaming_api.framing import frame_decode_all, frame_encode
 
 
 class SessionSettings(object):
 
     "Encapsulate settings necessary to create a new session"
 
-    def __init__(self, username, password):
-        if username is None:
-            raise ValueError("username cannot be None")
-        if password is None:
-            raise ValueError("password cannot be None")
+    def __init__(self, username, password, host='api-sandbox.smarkets.com', port=3801, ssl=True,
+                 socket_timeout=30, ssl_kwargs=None):
         self.username = username
         self.password = password
-        self.host = 'api-sandbox.smarkets.com'
-        self.port = 3701
-        self.socket_timeout = 30
-        self.ssl = True
-        self.ssl_kwargs = {}
+        self.host = host
+        self.port = port
+        self.socket_timeout = socket_timeout
+        self.ssl = ssl
+        self.ssl_kwargs = ssl_kwargs or {}
         # Most message are quite small, so this won't come into
         # effect. For larger messages, it needs some performance
         # testing to determine whether a single large recv() system
         # call is worse than many smaller ones.
         self.read_chunksize = 65536  # 64k
 
-    def validate(self):
-        "Do basic validation on the settings"
-        if not isinstance(self.socket_timeout, (types.NoneType, int)):
-            raise ValueError("socket_timeout must be an integer or None")
-        if self.socket_timeout is not None and self.socket_timeout <= 0:
-            raise ValueError("socket_timeout must be positive")
-        if not isinstance(self.read_chunksize, (types.NoneType, int, long)):
-            raise ValueError("read_chunksize must be an integer or None")
-        if self.read_chunksize is not None and self.read_chunksize <= 0:
-            raise ValueError("read_chunksize must be positive")
-        if not isinstance(self.ssl, (bool)):
-            raise ValueError("ssl must ba a bool")
-        if not isinstance(self.ssl_kwargs, (dict)):
-            raise ValueError("ssl_kwargs must be a dict")
-
 
 class Session(object):
 
@@ -63,9 +44,9 @@ class Session(object):
     flush_logger = private(logging.getLogger('smarkets.session.flush'))
 
     def __init__(self, settings, inseq=1, outseq=1, account_sequence=None):
-        if not isinstance(settings, SessionSettings):
-            raise ValueError("settings is not a SessionSettings")
-        settings.validate()
+        """
+        :type setting: :class:`SessionSettings`
+        """
         self.settings = settings
         self.account_sequence = account_sequence
         self.socket = SessionSocket(settings)
@@ -76,7 +57,9 @@ class Session(object):
         self.buf_outseq = outseq
         self.in_payload = seto.Payload()
         self.out_payload = seto.Payload()
-        self.send_buffer = Queue.Queue()
+        self.send_buffer = b''
+        self.read_buffer = b''
+        self.buffered_incoming_payloads = []
 
     @property
     def raw_socket(self):
@@ -86,6 +69,10 @@ class Session(object):
         '''
         return self.socket._sock
 
+    @property
+    def output_buffer_size(self):
+        return len(self.send_buffer)
+
     @property
     def connected(self):
         "Returns True if the socket is currently connected"
@@ -94,10 +81,9 @@ class Session(object):
     def connect(self):
         "Connects to the API and logs in if not already connected"
         if self.socket.connect():
-            # Clear the outgoing send buffer
-            self.send_buffer = Queue.Queue()
+            self._clear_send_buffer()
             # Reset separate outgoing buffer sequence number
-            self.buf_outseq = self.outseq
+            self.buf_outseq = 1
             login = self.out_payload
             login.Clear()
             login.type = seto.PAYLOAD_LOGIN
@@ -110,8 +96,16 @@ class Session(object):
                                  self.account_sequence)
                 login.login.account_sequence = self.account_sequence
 
-            # Always flush outgoing login message
-            self.send(True)
+            self.send()
+            self.flush()
+
+    def _clear_send_buffer(self):
+        if self.send_buffer:
+            self.logger.warn(
+                'Clearing non-empty buffer, %d bytes will be lost: %r',
+                len(self.send_buffer), self.send_buffer,
+            )
+        self.send_buffer = b''
 
     def logout(self):
         "Disconnects from the API"
@@ -121,65 +115,67 @@ class Session(object):
         logout.eto_payload.type = eto.PAYLOAD_LOGOUT
         logout.eto_payload.logout.reason = eto.LOGOUT_NONE
         self.logger.info("sending logout payload")
-        self.send(True)
+        self.send()
+        self.flush()
 
     def disconnect(self):
         "Disconnects from the API"
         self.socket.disconnect()
 
-    def send(self, flush=False):
+    def send(self):
         "Serialise, sequence, add header, and send payload"
         self.logger.debug(
-            "buffering payload with outgoing sequence %d: %s",
-            self.outseq, LazyCall(MessageToString, self.out_payload))
+            "buffering payload: %s",
+            LazyCall(MessageToString, self.out_payload))
         sent_seq = self.buf_outseq
         self.out_payload.eto_payload.seq = sent_seq
-        self.send_buffer.put_nowait(self.out_payload.SerializeToString())
+        self.send_buffer += frame_encode(self.out_payload.SerializeToString())
         self.buf_outseq += 1
-        if flush:
-            self.flush()
 
     def flush(self):
         "Flush payloads to the socket"
-        self.flush_logger.debug("flushing %d payloads", self.send_buffer.qsize())
-        while not self.send_buffer.empty():
-            try:
-                msg_bytes = self.send_buffer.get_nowait()
-                self.socket.send(msg_bytes)
-                self.outseq += 1
-            except Queue.Empty:
-                break
+        self.flush_logger.debug("Flushing %d bytes", len(self.send_buffer))
+        if self.send_buffer:
+            bytes_sent = self.socket.send(self.send_buffer)
+            self.flush_logger.debug("Flushed %d bytes out of %d", bytes_sent, len(self.send_buffer))
+            self.send_buffer = self.send_buffer[bytes_sent:]
+
+    def read(self):
+        self.read_buffer += self.socket.recv()
+
+        messages, remaining_buffer = frame_decode_all(self.read_buffer)
+        self.buffered_incoming_payloads.extend(messages)
+        self.read_buffer = remaining_buffer
 
     def next_frame(self):
-        "Get the next frame and increment inseq"
-        msg_bytes = self.socket.recv()
+        """Get the next payload and increment inseq.
+
+        .. warning::
+            Payload returned by `next_frame` has to be consumed before next call to `next_frame` happens.
+
+        :return: A payload or None if no payloads in buffer.
+
+        :rtype: :class:`smarkets.streaming_api.seto.Payload` or None
+
+        """
+        if not self.buffered_incoming_payloads:
+            return None
+
+        data, self.buffered_incoming_payloads = (
+            self.buffered_incoming_payloads[0], self.buffered_incoming_payloads[1:])
+
         self.in_payload.Clear()
-        self.in_payload.ParseFromString(msg_bytes)
+        self.in_payload.ParseFromString(data)
         self._handle_in_payload()
         if self.in_payload.eto_payload.seq == self.inseq:
             # Go ahead
             self.logger.debug("received sequence %d", self.inseq)
             self.inseq += 1
             return self.in_payload
-        elif self.in_payload.eto_payload.type == eto.PAYLOAD_REPLAY:
-            # Just a replay message, sequence not important
-            seq = self.in_payload.eto_payload.replay.seq
-            self.logger.debug("received a replay message with sequence %d", seq)
-            return None
         elif self.in_payload.eto_payload.seq > self.inseq:
-            # Need a replay
-            self.logger.info(
-                "received incoming sequence %d, expected %d, need replay",
-                self.in_payload.eto_payload.seq,
-                self.inseq)
-            replay = self.out_payload
-            replay.Clear()
-            replay.type = seto.PAYLOAD_ETO
-            replay.eto_payload.type = eto.PAYLOAD_REPLAY
-            replay.eto_payload.replay.seq = self.inseq
-            # Do not auto-flush replay because we may be in another
-            # thread
-            self.send()
+            self.logger.warn(
+                'Received incoming sequence %d instead of expected %d',
+                self.in_payload.eto_payload.seq, self.inseq)
             return None
         else:
             return None
@@ -190,19 +186,16 @@ class Session(object):
         self.logger.debug("received message to dispatch: %s", LazyCall(MessageToString, msg))
         if msg.eto_payload.type == eto.PAYLOAD_LOGIN_RESPONSE:
             self.session = msg.eto_payload.login_response.session
-            self.outseq = msg.eto_payload.login_response.reset
-            self.buf_outseq = self.outseq
-            self.send_buffer = Queue.Queue()
+            self.buf_outseq = msg.eto_payload.login_response.reset
+            self._clear_send_buffer()
             self.logger.info("received login_response with session %r and outseq %d",
-                             self.session, self.outseq)
+                             self.session, self.buf_outseq)
         elif msg.eto_payload.type == eto.PAYLOAD_HEARTBEAT:
             self.logger.debug("received heartbeat message, responding...")
             heartbeat = self.out_payload
             heartbeat.Clear()
             heartbeat.type = seto.PAYLOAD_ETO
             heartbeat.eto_payload.type = eto.PAYLOAD_HEARTBEAT
-            # Do not immediately flush heartbeat response because we
-            # may be in another thread
             self.send()
         return msg
 
@@ -217,7 +210,6 @@ class SessionSocket(object):
         if not isinstance(settings, SessionSettings):
             raise ValueError("settings is not a SessionSettings")
         self.settings = settings
-        self._buffer = ''
         self._sock = None
 
     @property
@@ -267,101 +259,42 @@ class SessionSocket(object):
             self.logger.debug('Exception when disconnecting: %r', e)
         self._sock = None
 
-    def send(self, msg_bytes):
-        "Send a payload"
+    def send(self, data):
+        """
+        :type data: byte string
+        :return: Number of sent bytes
+        :rtype: int
+        """
         if self._sock is None:
             raise SocketDisconnected('Trying to write to socket when disconnected')
-        byte_count = len(msg_bytes)
-        # Pad to 4 bytes
-        padding = '\x00' * max(0, 3 - byte_count)
-        self.logger.debug(
-            "payload has %d bytes and needs %d padding",
-            byte_count, len(padding))
-        frame = _encode_varint(byte_count) + msg_bytes + padding
         try:
-            self.wire_logger.debug("sending frame bytes %r", frame)
-            self._sock.sendall(frame)
+            self.wire_logger.debug("sending %d bytes: %r", len(data), data)
+            sent = self._sock.send(data)
+            if sent == 0:
+                raise SocketDisconnected('Socket disconnected when writing to it, 0 bytes written')
+            return sent
         except socket.error as e:
             reraise(ConnectionError("Error while writing to socket", e))
 
     def recv(self):
-        "Read a frame with header"
-        # Read a minimum of 4 bytes
-        self._fill_buffer()
-        result = 0
-        shift = 0
-        pos = 0
-        while 1:
-            if pos > len(self._buffer) - 1:
-                pos = 0
-                # Empty buffer and read another 4 bytes
-                self._fill_buffer(4, True)
-            cbit = ord(self._buffer[pos])
-            result |= ((cbit & 0x7f) << shift)
-            pos += 1
-            if not (cbit & 0x80):
-                self._buffer = self._buffer[pos:]
-                to_read = max(0, result - len(self._buffer))
-                self.logger.debug("next message is %d bytes long", to_read)
-                if to_read:
-                    # Read the actual message if necessary
-                    while to_read > self.settings.read_chunksize:
-                        self._fill_buffer(
-                            self.settings.read_chunksize + len(self._buffer))
-                        to_read -= self.settings.read_chunksize
-                    if to_read > 0:
-                        self._fill_buffer(to_read + len(self._buffer))
-                msg_bytes = self._buffer[:result]
-                self.wire_logger.debug("received bytes %r", msg_bytes)
-                # Consume the buffer
-                self._buffer = self._buffer[result:]
-                return msg_bytes
-            shift += 7
+        """Read stuff from underlying socket.
 
-    def _fill_buffer(self, min_size=4, empty=False):
-        "Ensure the buffer has at least 4 bytes"
+        :rtype: byte string
+        """
         if self._sock is None:
             raise SocketDisconnected(
                 'Trying to read from a socket when disconnected')
 
-        if empty:
-            self._buffer = ''
-        while len(self._buffer) < min_size:
-            bytes_needed = min_size - len(self._buffer)
-            self.logger.debug("receiving %d bytes", bytes_needed)
-            try:
-                if self.settings.ssl:
-                    inbytes = self._recv_msg_ssl(bytes_needed)
-                else:
-                    inbytes = self._recv_msg_sock(bytes_needed)
-            except socket.error as e:
-                reraise(ConnectionError('Error while reading from socket', e))
-
-            self._buffer += inbytes
-
-    def _recv_msg_sock(self, bytes_needed):
-        "Wrap reading from a plain socket"
-        inbytes = self._sock.recv(bytes_needed, socket.MSG_WAITALL)
-        if len(inbytes) != bytes_needed:
-            message = "Socket disconnected while receiving, got %r" % (inbytes,)
-            self.logger.info(message)
-            raise SocketDisconnected(message)
-        return inbytes
-
-    def _recv_msg_ssl(self, bytes_needed):
-        "Wrap reading from an SSL socket where we can't pass options to recv"
-        msglen = 0
-        msglist = []
-        while msglen < bytes_needed:
-            chunk = self._sock.recv(
-                min(self.settings.read_chunksize, bytes_needed - msglen))
-            if not chunk:
-                message = "Socket disconnected while receiving"
+        try:
+            inbytes = self._sock.recv(self.settings.read_chunksize)
+            if not inbytes:
+                message = "Socket disconnected while receiving, got %r" % (inbytes,)
                 self.logger.info(message)
                 raise SocketDisconnected(message)
-            msglist.append(chunk)
-            msglen += len(chunk)
-        return ''.join(msglist)
+            self.wire_logger.debug('Received %d bytes: %r', len(inbytes), inbytes)
+            return inbytes
+        except socket.error as e:
+            reraise(ConnectionError('Error while reading from socket', e))
 
     def _error_message(self, exception):
         "Stringify a socket exception"
@@ -374,17 +307,3 @@ class SessionSocket(object):
             return "Error %s connecting %s:%s. %s." % (
                 exception.args[0], self.settings.host, self.settings.port,
                 exception.args[1])
-
-
-def _encode_varint(value):
-    """
-    Encode an int/long as a ULEB128 number
-    """
-    bits = value & 0x7f
-    value >>= 7
-    ret = ''
-    while value:
-        ret += chr(0x80 | bits)
-        bits = value & 0x7f
-        value >>= 7
-    return ret + chr(bits)
